from __future__ import annotations

import re
import qrcode
from pathlib import Path


RE_QR = re.compile(r'\!\[qr\:(?P<alt>[^\n\s]+)\]\((?P<text>[^\n\s]+)\)')


class QRGenerator:
    def __init__(self, save_dir: Path, reference_dir: Path):
        self.save_dir = Path(save_dir)
        self.reference_dir = Path(reference_dir)
        self.image_index = 0

    @classmethod
    def from_paths(cls, paths: str) -> QRGenerator:
        """
        Split string into two paths
            1. Where to save generated QR codes
            2. Directory to reference the images
        """
        save_dir, reference_dir = paths.split(':')

        save_dir = Path(save_dir)
        reference_dir = Path(reference_dir)

        return QRGenerator(save_dir, reference_dir)

    def repl_qr(self, match: re.Match) -> str:
        text, alt = match.group('text', 'alt')

        # Get image name by index
        image_name = f'qr-{self.image_index}-{alt}.png'
        self.image_index += 1
        
        # Generate image from text
        image = qrcode.make(text)

        # Save image
        self.save_dir.mkdir(parents=True, exist_ok=True)
        image.save(self.save_dir / image_name)

        # Reference image and return HTML
        image_path = self.reference_dir / image_name
        return f'<img src="{image_path!s}" alt="{alt}">'


def qr(document: str, qrgen: QRGenerator) -> str:
    """
    Notation for QR codes

    QR code images will be generated and saved,
    and image elements inserted referencing their file locations.


    Example:
        Saving images to `public/imgs/qr`
        Referencing them relative to `/imgs/qr`

        Marquedown:
            ![qr:monero](monero:abcdefhijklmnopqrstuvwxyz)
            ![qr:dimweb](https://youtu.be/VmAEkV5AYSQ)

        HTML:
            <img src="/imgs/qr/qr-0.png" alt="monero">
            <img src="/imgs/qr/qr-1.png" alt="dimweb">
    """

    document = RE_QR.sub(qrgen.repl_qr, document)

    return document


import re

from . import marquedown as mqd
from .tagtools import tagstrip, contained


RE_LABELLIST = re.compile(r'^(?:(?P<class>[^\s\n]+)\:\n)?(?P<labels>(?:\(\|[ ]+[\w\d\-\_\:\.]+\:[ ]+.+(?:\n|$))+)', re.MULTILINE)
RE_LABEL = re.compile(r'^\(\|[ ]+([\w\d\-\_\:\.]+)\:[ ]+(.+)$')


def _repl_labellist(match: re.Match):
    # Extract label name and content from each line
    labels = []
    
    for line in match.group('labels').splitlines():
        m = RE_LABEL.match(line)

        if not m:
            raise RuntimeError(
                'Line in label list is not a label. '
                'This error should be impossible to occur.')

        name, content = m.group(1, 2)

        # Parse Marquedown in content
        parsed_content = mqd(content)
        
        # Strip added paragraph tags
        if not contained(content, 'p'):
            parsed_content = tagstrip(parsed_content, 'p')

        labels.append((name, parsed_content))

    # Render list items
    rendered_labels = "\n".join(
        f'<li class="label label-{name}">{content}</li>'
        for name, content in labels)

    # And finally, render the label list

    _class = match.group('class')
    if _class is not None:
        # Add class to list if one is provided
        return f'<ul class="labels {_class}">\n{rendered_labels}\n</ul>'
    else:
        return f'<ul class="labels">\n{rendered_labels}\n</ul>'


def labellist(document: str):
    """Notation for label lists.
    
    Marquedown:
        (| email: [jon@webby.net](mailto:jon@webby.net)
        (| matrix: [@jon:webby.net](https://matrix.to/#/@jon:webby.net)
        (| runescape: jonathan_superstar1777

    HTML:
        <ul class="labels">
            <li class="label-email">
                <a href="mailto:jon@webby.net">
                    jon@webby.net
                </a>
            </li>
            <li class="label-matrix">
                <a href="https://matrix.to/#/@jon:webby.net">
                    @jon:webby.net
                </a>
            </li>
            <li class="label-runescape">
                jonathan_superstar1777
            </li>
        </ul>
    """

    return RE_LABELLIST.sub(_repl_labellist, document)


import re


RE_YOUTUBE_VIDEO = re.compile(r'\!\[.+\]\(https\:\/\/(?:(?:www\.)?youtube\.com|youtu\.be)\/([0-9A-Za-z-_]+)(?: \"(.*)\")?\)')


def _repl_youtube_video(match: re.Match):
    video_id, title = match.group(1, 2)
    return (
        f'<iframe '
        f'class="mqd-video" '
        f'src="https://www.youtube.com/embed/{video_id}" '
        f'title={title or "" !r} '
        f'frameborder="0" '
        f'allow="accelerometer; autoplay; clipboard-write; encrypted-media; gyroscope; picture-in-picture" '
        f'allowfullscreen>'
        f'</iframe>'
    )


def video(document: str):
    """Notation for videos from recognised
    video sharing platforms that allow embedding.

    Supported platforms: YouTube
    
    YouTube:
        Marquedown:
            ![dimweb](https://youtu.be/VmAEkV5AYSQ "An embedded YouTube video")
        
        HTML:
            <iframe
                class="mqd-video"
                src="https://www.youtube.com/embed/VmAEkV5AYSQ"
                title="An embedded YouTube video" frameborder="0"
                allow="accelerometer; autoplay; clipboard-write; encrypted-media; gyroscope; picture-in-picture"
                allowfullscreen>
            </iframe>
    """

    # Embed all YouTube videos
    document = RE_YOUTUBE_VIDEO.sub(_repl_youtube_video, document)

    return document


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['marquedown', 'marquedown.commands']

package_data = \
{'': ['*']}

install_requires = \
['Markdown>=3.3.6,<4.0.0', 'PyYAML>=6.0,<7.0']

setup_kwargs = {
    'name': 'marquedown',
    'version': '0.9.0',
    'description': 'Extending Markdown further by adding a few more useful notations.',
    'long_description': '# Marquedown\n\nExtending Markdown further by adding a few more useful notations.\nIt can be used in place of `markdown` as it also uses and applies it.\n\n## Examples\n\n### Blockquote with citation\n\nThis is currently limited to the top scope with no indentation.\nSurrounding dotted lines are optional.\n\n#### Example\n\n##### Marquedown\n\n```md\n......................................................\n> You have enemies? Good. That means you\'ve stood up\n> for something, sometime in your life.\n-- Winston Churchill\n\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\n```\n\n##### HTML\n\n```html\n<blockquote>\n    <p>\n        You have enemies? Good. That means you\'ve stood up\n        for something, sometime in your life.\n    </p>\n    <cite>Winston Churchill</cite>\n</blockquote>\n```\n\n### Embed video\n\n#### YouTube\n\n##### Marquedown\n\n```md\n![dimweb](https://youtu.be/VmAEkV5AYSQ "An embedded YouTube video")\n```\n\n##### HTML\n\n```html\n<iframe\n    src="https://www.youtube.com/embed/VmAEkV5AYSQ"\n    title="An embedded YouTube video" frameborder="0"\n    allow="accelerometer; autoplay; clipboard-write; encrypted-media; gyroscope; picture-in-picture"\n    allowfullscreen>\n</iframe>\n```\n\n### BBCode HTML tags\n\nThese tags allow you to put Marquedown inside HTML tags. This is done by finding and replacing them with their represented HTML after all other Marquedown has been rendered.\n\n#### Tags and classes\n\nThe naming scheme is the same as in CSS, e.g. `tag.class1.class2`\nIf `tag` is omitted, it is treated to be `div`\n\n#### ID:s\n\nID:s are supported using `#beans` at the end of the tag, ex. `[p#beans]`\n\n#### Example\n\n##### Marquedown\n\n```md\n[section]\n[.bingo]\nA regular **paragraph** written in Marquedown, but *within* other HTML tags.\n[//]\n\n[tag.class1.class2] [/tag]\n```\n\n##### HTML\n\n```html\n<section>\n<div class="bingo">\n    <p>\n        A regular <strong>paragraph</strong> written in Marquedown, but <em>within</em> other HTML tags.\n    </p>\n</div></section>\n\n<tag class="class1 class2"> </tag>\n```\n\n### Label list\n\n#### Example\n\n##### Marquedown\n\n```md\n(| email: [jon@webby.net](mailto:jon@webby.net)\n(| matrix: [@jon:webby.net](https://matrix.to/#/@jon:webby.net)\n(| runescape: jonathan_superstar1777\n```\n\n##### HTML\n\n```html\n<ul class="labels">\n    <li class="label label-email">\n        <a href="mailto:jon@webby.net">\n            jon@webby.net\n        </a>\n    </li>\n    <li class="label label-matrix">\n        <a href="https://matrix.to/#/@jon:webby.net">\n            @jon:webby.net\n        </a>\n    </li>\n    <li class="label label-runescape">\n        jonathan_superstar1777\n    </li>\n</ul>\n```\n\n### QR codes\n\nQR codes can be generated and referenced automatically via the `render` command. If you want to generate QR codes with `marquedown.marquedown`, you can follow the example below.\n\n#### Example\n\n##### Marquedown\n\n```md\n![qr:monero-wallet](monero:abcdefghijklmnopqrstuvwxyz)\n```\n\n##### Python\n\n```py\nfrom marquedown import marquedown\n\nwith open(\'document.mqd\', \'r\') as f:\n    document = f.read()\n\nwith open(\'public/document.html\', \'w\') as f:\n    rendered = marquedown(document, qrgen=\'public/qr:qr\')\n    f.write(rendered)\n```\n\nThe string provided to `qrgen` is two paths separated by a colon.\n\n1. Where to save the generated images\n2. What directory to reference in the HTML `src` parameter\n\n##### The generated HTML\n\n```html\n<img src="qr/qr-0-monero-wallet.png" alt="monero-wallet">\n```\n\n## Commands\n\n### `render`: Render documents\n\nYou can render an entire directory and its subdirectories of Markdown or Marquedown documents. This can be used to automate rendering pages for your website.\n\nDo `python -m marquedown render --help` for list of options.\n\n#### Example\n\nFor a few of my websites hosted on GitLab, I have it set up to run *this* on push:\n\n```sh\n# Render document\npython -m marquedown render -i "./mqd" -o "./public" -t "./templates/page.html"\n\n# This is for the GitLab Pages publication\nmkdir .public\ncp -r public .public\nmv .public public  \n```',
    'author': 'Maximillian Strand',
    'author_email': 'maximillian.strand@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/deepadmax/marquedown',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



# Render all Markdown and Marquedown documents
# from a specified directory and all its subdirectories.
# 
# Provide a template for embedding the rendered content.

import re
import os
import yaml

from pathlib import Path
from collections import defaultdict

from .. import marquedown
from ..qr import QRGenerator


def extract_metadata(document: str):
    """Extract metadata from a YAML code block at the top of the document.
    
    Example:
        ```yml
        title: Page Title
        ```
    """

    m = re.search(r'^\`{3}ya?ml\n((?:.|\n)*?)\n\`{3}\s+', document)

    # Return without metadata if missing
    if not m:
        return {}, document

    # Remove metadata from document
    document = document[len(m.group(0)):]

    # Attempt to load metadata from YAML
    metadata = yaml.safe_load(m.group(1)) or {}

    if not isinstance(metadata, dict):
        raise ValueError('Metadata must be a dictionary.')

    return metadata, document


def find_documents(source: Path, destination: Path):
    """Locate all Marquedown documents in the source directory and its subdirectories.
    Yields the source of every document, and the destination for its `index.html`.
    """

    for directory, _, filenames in os.walk(source):
        for fname in filenames:
            file_source = Path(directory) / fname

            if not file_source.is_file():
                continue
            if file_source.suffix not in ('.md', '.mqd'):
                continue

            # File destination from relative path without suffix
            file_relative_path = file_source.relative_to(source)
            file_relative_path = str(file_relative_path)[:-len(file_relative_path.suffix)]
            file_destination = destination / file_relative_path

            # Don't use document name if it is `index.md` or `index.mqd`
            # This should be equivalent to how `index.html` works and be hidden

            if file_source.name in ('index.md', 'index.mqd'):
                yield file_source, file_destination.parent / 'index.html'
            else:
                yield file_source, file_destination / 'index.html'


def render_documents(source: Path, destination: Path, template: Path = None):
    """
    Load all documents from within the source directory
    and its subdirectories and render them onto a template file.
    Save all renders in the destination directory.
    """

    # Load template
    if template is not None:
        with open(template, 'r') as f:
            template = f.read()

    # Iterate over the source and destination directories for each document
    for fsource, fdestination in find_documents(source, destination):
        # Load document
        with open(fsource, 'r') as f:
            document = f.read()

        # Prepare document and render it
        metadata, document = extract_metadata(document)

        # Ensure destination directory
        os.makedirs(fdestination.parent, exist_ok=True)

        # Save QR codes differently depending on whether source file is index
        if fsource.name in ('index.md', 'index.mqd'):
            qrgen = QRGenerator(fdestination.parent / 'qr', Path('qr'))
        else:
            qrgen = QRGenerator(fdestination.parent / 'qr', Path(fsource.stem, 'qr'))


        # Render Marquedown, optionally template, and write to file
        rendered = marquedown(document, qrgen=qrgen)

        with open(fdestination, 'w') as f:
            # Format to template if provided
            if template is not None:
                rendered = template.format_map(defaultdict(str, **metadata, document=rendered))
            # Write to destination
            f.write(rendered)
            print(f'Rendered document: {fsource.relative_to(source)}')


import re

from . import marquedown as mqd


RE_CITATION = re.compile(r'^(?:\.+\n)?((?:\>.*\n)+)(?:\n?\-{2}[ ](.+)\n?)?(?:\'+)?', flags=re.MULTILINE)


def _repl_citation(match: re.Match):
    quote, source = match.group(1, 2)

    # Remove angle brackets in quote
    quote = '\n'.join(line[1:].lstrip() for line in quote.splitlines())

    # Parse Marquedown in quote
    quote = mqd(quote)

    # Include source if provided
    if source is not None:
        # Parse Marquedown in source
        source = mqd(source)

        # Remove surrounding paragraph from source
        if source.startswith('<p>') and source.endswith('</p>'):
            source = source[3:-4]

        # Put everything into HTML
        return f'<blockquote>\n{quote}\n<cite>{source}</cite>\n</blockquote>'

    # Exclude source if not provided
    return f'<blockquote>\n{quote}\n</blockquote>'


def citation(document: str):
    """Notation for blockquotes that include citation.

    Marquedown:
        ......................................................
        > You have enemies? Good. That means you've stood up
        > for something, sometime in your life.
        -- Winston Churchill
        ''''''''''''''''''''''''''''''''''''''''''''''''''''''

    HTML:
        <blockquote>
            <p>
                You have enemies? Good. That means you've stood up
                for something, sometime in your life.
            </p>
            <cite>Winston Churchill</cite>
        </blockquote>
    """

    return RE_CITATION.sub(_repl_citation, document)